// ---------------------------------------------------------------------------------------

// Gira Funkbus Interface by EL
// Software to capture and send Funkbus messages
// see file : FunkBus protocol Berker.txt
// Hardware interface, see pictures of schematics and experimental hardware
// Version 01-05-2024

// ---------------------------------------------------------------------------------------

#include <stdio.h>
#include <cstdio>
#include <string.h>

#include "EL_FunkBus.hpp"

// ---------------------------------------------------------------------------------------
// FreeRtos
// ---------------------------------------------------------------------------------------

#ifndef CONFIG_FREERTOS_HZ
#define CONFIG_FREERTOS_HZ 100
#endif
#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "freertos/queue.h"

// ---------------------------------------------------------------------------------------
// Hardware interface
// ---------------------------------------------------------------------------------------

#include "driver/gpio.h"
#include "esp_timer.h"

// ---------------------------------------------------------------------------------------
// Debug help
// ---------------------------------------------------------------------------------------

#include "esp_err.h"
#include "esp_log.h"
static const char *TAG = "FunkBus";

// ---------------------------------------------------------------------------------------
// Bit time limits funkbus protocol
// ---------------------------------------------------------------------------------------

#define BIT1_MIN 350
#define BIT1_MAX 650
#define BIT0_MIN 850
#define BIT0_MAX 1200

// ---------------------------------------------------------------------------------------
// FreeRtos queues
// ---------------------------------------------------------------------------------------

QueueHandle_t funkBusReceiveQueue;
#define FUNKBUS_RECEIVE_QUEUE_LEN 64

QueueHandle_t funkBusBitSendQueue;
#define FUNKBUS_BITSEND_QUEUE_LEN 64

QueueHandle_t funkBustelegramRawSendQueue;
#define FUNKBUS_MEMSEND_QUEUE_LEN 64

// ---------------------------------------------------------------------------------------
// Definition Raw funkbus data
// ---------------------------------------------------------------------------------------

#define TELEGRAM_BYTECOUNT 6

typedef struct 
{
    uint8_t raw[TELEGRAM_BYTECOUNT];
} funkBusTelegramRaw;

// ---------------------------------------------------------------------------------------
// Receive FunkBus edges interrupt
// ---------------------------------------------------------------------------------------

volatile static int64_t oldTimeStamp = 0;

/// @brief Receive interrupt handler receiving receive-pin transitions.
/// @param arg See ESP-IDF manual.
static void IRAM_ATTR funkBusReceiveInteruptHandler(void *arg)
{
    int64_t timeStamp = esp_timer_get_time();
    int64_t deltaTime = timeStamp - oldTimeStamp;
    if (deltaTime >= BIT1_MIN)
    {
        if (deltaTime <= BIT0_MAX)
        {
            xQueueSendFromISR(funkBusReceiveQueue, &deltaTime, NULL);
        }
        else
        {
            if (deltaTime > BIT0_MAX)
            {
                deltaTime = BIT0_MAX + 1;
                xQueueSendFromISR(funkBusReceiveQueue, &deltaTime, NULL);
            }
        }
        oldTimeStamp = timeStamp;
    }
}

// ---------------------------------------------------------------------------------------
// FunkBus data info
// ---------------------------------------------------------------------------------------

#define TRANSMITTER_TYPE            0x2

#define TRANSMITTER_SUBTYPE         0xC

#define TRANSMITTER_STATE           0x5

#define TELEGRAM_KEY_BITMASK        0x7
#define TELEGRAM_CHANNEL_BITMASK    0x3
#define TELEGRAM_SEQUENCE_BITMASK   0x3

#define BIT_STRING_LEN_MAX          ((TELEGRAM_BYTECOUNT * 8) - 1)

#define TELEGRAM_BYTE_WITH_CS       5
#define TELEGRAM_BYTE_WITH_DATA_CB  5

// ---------------------------------------------------------------------------------------
// Send timer handler
// ---------------------------------------------------------------------------------------

// definition for the sending hardware.
#define SEND_PIN_RELEASE    0   // false
#define SEND_PIN_MARK       1   // true

// memmory to identify used hardware
esp_timer_handle_t funkBusBitSentTimer;
gpio_num_t rxPin = GPIO_NUM_NC;
gpio_num_t txPin = GPIO_NUM_NC;

/// @brief Handler called by timer to send bits on the queue
/// @param arg 
static void funkBusSendingTimerHandler(void* arg)
{
    static bool funkBusBitIsSending = false;
    static bool funkBusSending1 = false;
    static bool funkBusState = SEND_PIN_RELEASE;
    static uint8_t postAmble = 0;
    gpio_set_level(txPin, funkBusState);
    if (funkBusBitIsSending)
    {
        if (funkBusSending1)
        {
            funkBusState = !funkBusState;
        }
        funkBusBitIsSending = false;
    }
    else
    {
        bool bitToSend;
        if (xQueueReceiveFromISR(funkBusBitSendQueue, &bitToSend, 0) == pdTRUE)
        {
            funkBusSending1 = bitToSend;
            funkBusBitIsSending = true;
            funkBusState = !funkBusState;
            postAmble = 10;
        }
        else
        {
            if (postAmble > 0)
            {
                funkBusState = true;
                postAmble--;
            }
            else
            {
                esp_timer_stop(funkBusBitSentTimer);
                funkBusBitIsSending = false;
                funkBusState = SEND_PIN_RELEASE; // When idle, do not drive bus
                gpio_set_level(txPin, funkBusState);
            }
        }
    }
}

// ---------------------------------------------------------------------------------------
// Set hardware FunkBus receive pin, and interrupt system
// ---------------------------------------------------------------------------------------

/// @brief Function to initialise the receiving hardware (rx-pin) and install the receive interrupt handler
void setFunkBusReceiveHardware()
{
    gpio_config_t busPin = {};
    busPin.pin_bit_mask = 1ULL << rxPin;
    busPin.intr_type = GPIO_INTR_ANYEDGE; // Funkbus protocol apears to have no polarity only interval time is relevant.
    busPin.mode = GPIO_MODE_INPUT;
    busPin.pull_down_en = GPIO_PULLDOWN_ENABLE;
    gpio_config(&busPin);
    funkBusReceiveQueue = xQueueCreate(FUNKBUS_RECEIVE_QUEUE_LEN, sizeof(int64_t));
    if (gpio_install_isr_service(ESP_INTR_FLAG_LEVEL2) != ESP_OK)
    {
        ESP_LOGE(TAG, "Cannot install isr service");
    }
    if (gpio_isr_handler_add(rxPin, funkBusReceiveInteruptHandler, (void *)rxPin) != ESP_OK)
    {
        ESP_LOGE(TAG, "Cannot add isr handler");
    }
}

// ---------------------------------------------------------------------------------------
// Set hardware FunkBus send pin, and timer system
// ---------------------------------------------------------------------------------------

/// @brief Function to initialise hardware to send. (timer and tx-pin)
void setFunkBusSendHardware()
{
    gpio_config_t io_pin = {};
    io_pin.pin_bit_mask = 1ULL << txPin;
    io_pin.intr_type = GPIO_INTR_DISABLE;
    io_pin.mode = GPIO_MODE_OUTPUT;
    gpio_config(&io_pin);
    gpio_set_level(txPin, SEND_PIN_RELEASE);

    const esp_timer_create_args_t funkBusTimerArgs = 
    {
        .callback = &funkBusSendingTimerHandler,
        .arg = NULL,
        .dispatch_method = ESP_TIMER_TASK,
        .name = "FunkBusSend",
        .skip_unhandled_events = true
    };
    funkBusBitSendQueue = xQueueCreate(FUNKBUS_BITSEND_QUEUE_LEN, sizeof(bool));
    funkBusReceiveQueue = xQueueCreate(FUNKBUS_RECEIVE_QUEUE_LEN, sizeof(int64_t));
    funkBustelegramRawSendQueue = xQueueCreate(FUNKBUS_MEMSEND_QUEUE_LEN, sizeof(funkBusTelegramRaw));
    if (esp_timer_create(&funkBusTimerArgs, &funkBusBitSentTimer) != ESP_OK)
    {
        ESP_LOGE(TAG, "Cannot create timer");
    }
}

// ---------------------------------------------------------------------------------------
// telegramRaw -- checksuum -- parity
// ---------------------------------------------------------------------------------------

/// @brief Function to clear all raw telegram bits
/// @param telegramRaw Pointer to the raw telegram data
void funkBusCleartelegramRaw(funkBusTelegramRaw * telegramRaw)
{
    memset(telegramRaw->raw, 0, TELEGRAM_BYTECOUNT);
}

/// @brief Function to calculate the parity offset-value of the raw telegram data
/// @param telegramRaw Pointer to the raw telegram data
/// @return Returns the parity-offset (the difference between expected and all-0-bits)
uint8_t funkBusCalcCsTelegramRaw(const uint8_t * telegramRaw)
{
    int keys[] = {0xc, 0x2, 0x8, 0x3, 0xc, 0x2, 0x8, 0x3};
    int chk    = telegramRaw[TELEGRAM_BYTE_WITH_CS] & 0xf; // checksum in last nibble
    for (int j = 0; j < TELEGRAM_BYTECOUNT; ++j)
    {
        int x = telegramRaw[j];
        if (j == TELEGRAM_BYTE_WITH_CS)
        {
            x = x & 0xe0; // the last byte only has 3 bits, then parity and the checksum
        }
        // each bit MSB to LSB
        for (int k = 0; k < 8; ++k)
        {
            if (x & (0x80 >> k))
            {
                chk = chk ^ keys[k];
            }
        }
    }
    return chk;
}

/// @brief Function to check the parity bits in the raw-telegram data
/// @param telegramRaw Pointer to the raw telegram data
/// @return true if the parity is ok.
bool funkBusChecktelegramRaw(const uint8_t * telegramRaw)
{
    return (funkBusCalcCsTelegramRaw(telegramRaw) == 0);
}

/// @brief Function to check the data-parity of the data-part in the raw telegram.
/// @param telegramRaw Pointer to the (existing!) raw telegram structure
/// @return true when parity is ok
bool funkBusCalcDataCheckBit(const funkBusTelegramRaw * telegramRaw)
{
    bool parity = true;
    uint8_t bitCount = 0;
    const uint8_t * data = &telegramRaw->raw[0];
    uint8_t mask = 0x80;
    while (bitCount < 44)
    {
        if ((*data & mask) != 0)
        {
            parity = not parity;
        }
        mask = mask >> 1;
        if (mask == 0)
        {
            mask =0x80;
            data++;
        }
        bitCount++;
    }
    return parity;
}

// ---------------------------------------------------------------------------------------
// FunkBus telegram raw layout
// ---------------------------------------------------------------------------------------

/*

Type sUb       Serial      sTe        Data        Cs  End
ttttbbbb ssssssss ssssssss ttttdddd dddddddd ddddcccc e         <-- t=type b=subtype s=serial t=status d=data c=cs e=end
                                     kkkcc n fqqp               <-- k=key c=channel n=on f=off p=parity
                                              rl                <-- r=repeat l=long
                           uuuuuuuu u                           <-- unknown function
76543210 76543210 76543210 76543210 76543210 76543210 76543210  <-- bit numbers
00101100 01111001 11111010 01010000 01000001 01010011 1         <-- bitpattern example
0        1        2        3        4        5                  <-- byte nr

*/

// ---------------------------------------------------------------------------------------
// telegramRaw -- telegram functions
// ---------------------------------------------------------------------------------------

void copyBit(const uint8_t * const sourceByte, const uint8_t sourceBitNr, uint8_t * destByte, const uint8_t destBitNr)
{
    uint8_t maskedSource = *sourceByte & (1 << sourceBitNr);
    bool b = (maskedSource > 0);
    uint8_t destMask = (1 << destBitNr);
    *destByte = (*destByte & ~destMask);
    if (b)
    {
        *destByte = *destByte | destMask;
    }
}

void setBit(uint8_t * destByte, const uint8_t destBitNr)
{
    uint8_t destMask = (1 << destBitNr);
    *destByte = *destByte | destMask;
}

void setRepeatBit(funkBusTelegramRaw * telegramRaw)
{
    setBit(&telegramRaw->raw[5], 6);   // repeat-bit    
}

void setLongPressBit(funkBusTelegramRaw * telegramRaw)
{
    setBit(&telegramRaw->raw[5], 5);   // long-bit    
}

void setOnBit(funkBusTelegramRaw * telegramRaw)
{
    setBit(&telegramRaw->raw[4], 0);   // on-bit    
}

void setOffBit(funkBusTelegramRaw * telegramRaw)
{
    setBit(&telegramRaw->raw[5], 7);   // off-bit    
}

/// @brief Function to overwrite the checksum-bits in the raw telegram structure with a valid checksum.
/// @param telegramRaw Pointer to the (existing!) raw-telegram structure
void setChecksumsIntelegramRaw(funkBusTelegramRaw * telegramRaw)
{
    telegramRaw->raw[TELEGRAM_BYTE_WITH_DATA_CB] = (telegramRaw->raw[TELEGRAM_BYTE_WITH_DATA_CB] & 0xE0); // mask all parity bits
    if (not funkBusCalcDataCheckBit(telegramRaw))
    {
        telegramRaw->raw[TELEGRAM_BYTE_WITH_DATA_CB] = telegramRaw->raw[TELEGRAM_BYTE_WITH_DATA_CB] | 0x10;
    }
    telegramRaw->raw[TELEGRAM_BYTE_WITH_CS] = telegramRaw->raw[TELEGRAM_BYTE_WITH_CS] | (funkBusCalcCsTelegramRaw(&telegramRaw->raw[0]) & 0xF);
}

/// @brief Function to convert the telegram structure to raw data format
/// @param telegram pointer to the telgram to be converted
/// @param telegramRaw pointer to the (existing!) raw-telegram structure
void funkBusTelegramToTelegramRaw(const funkBusTelegram * telegram, funkBusTelegramRaw * telegramRaw)
{
    funkBusCleartelegramRaw(telegramRaw);
    telegramRaw->raw[0] = (TRANSMITTER_TYPE << 4) | (TRANSMITTER_SUBTYPE);
    telegramRaw->raw[1] = (telegram->button.serial >> 8);
    telegramRaw->raw[2] = (telegram->button.serial & 0x00FF);
    telegramRaw->raw[3] = (TRANSMITTER_STATE << 4); 
    uint8_t keyBinair = telegram->button.key - 1;
    copyBit(&keyBinair, 0, &telegramRaw->raw[4], 6);
    copyBit(&keyBinair, 1, &telegramRaw->raw[4], 5);
    copyBit(&keyBinair, 2, &telegramRaw->raw[4], 4);
    uint8_t channel = (uint8_t)telegram->channel;
    copyBit(&channel, 0, &telegramRaw->raw[4], 3);
    copyBit(&channel, 1, &telegramRaw->raw[4], 2);
    switch (telegram->type)
    {
        case noTask:
        break;
        case pressOff:
            setOffBit(telegramRaw);
        break;
        case pressOn:
            setOnBit(telegramRaw);
        break;
        case longPressOn:
            setOnBit(telegramRaw);
            setLongPressBit(telegramRaw);
        break;
        case longPressOff:
            setOffBit(telegramRaw);
            setLongPressBit(telegramRaw);
        break;
        case pressOnRepeat:
            setOnBit(telegramRaw);
            setRepeatBit(telegramRaw);
        break;
        case pressOffRepeat:
            setOffBit(telegramRaw);
            setRepeatBit(telegramRaw);
        break;
        case longPressOnRepeat:
            setOnBit(telegramRaw);
            setRepeatBit(telegramRaw);
            setLongPressBit(telegramRaw);
        break;
        case longPressOffRepeat:
            setOffBit(telegramRaw);
            setRepeatBit(telegramRaw);
            setLongPressBit(telegramRaw);
        break;
        case longPressEnd:
            setRepeatBit(telegramRaw);     // ??
            setLongPressBit(telegramRaw);
        break;
    }
    setChecksumsIntelegramRaw(telegramRaw);
}

uint16_t getSerialInfunkBustelegramRaw(const funkBusTelegramRaw * telegramRaw)
{
    return (telegramRaw->raw[1] << 8) | telegramRaw->raw[2];
}

funkBusChannel getChannelTypeInFunkBustelegramRaw(const funkBusTelegramRaw * telegramRaw)
{
    return (funkBusChannel)(( telegramRaw->raw[4] >> 2) & TELEGRAM_CHANNEL_BITMASK);
}

uint8_t getKeyInFunkBustelegramRaw(const funkBusTelegramRaw * telegramRaw)
{
    uint8_t i = 1;  // keys count from 1
    if ((telegramRaw->raw[4] & 0x40) > 0)
    {
        i = i + 1;    
    }
    if ((telegramRaw->raw[4] & 0x20) > 0)
    {
        i = i + 2;    
    }
    if ((telegramRaw->raw[4] & 0x10) > 0)
    {
        i = i + 4;    
    }
    return i;
}

uint8_t getSequenceInFunkBustelegramRaw(const funkBusTelegramRaw * telegramRaw)
{
    return (telegramRaw->raw[5] >> 5) & TELEGRAM_SEQUENCE_BITMASK;
}

bool getOnBitInFunkBustelegramRaw(const funkBusTelegramRaw * telegramRaw)
{
    return (telegramRaw->raw[4] & 0x01) > 0;
}

bool getOffBitInFunkBustelegramRaw(const funkBusTelegramRaw * telegramRaw)
{
    return (telegramRaw->raw[5] & 0x80) > 0;
}

bool getRepeatBitIntelegramRaw(const funkBusTelegramRaw * telegramRaw)
{
    return (telegramRaw->raw[5] & 0x40) > 0;
}

telegramType getTaskType(const bool onBit, const bool offBit, const uint8_t sequence)
{
    switch (sequence)
    {
        case 0b00:
            if (onBit)
            {
                return pressOn;
            }
            if (offBit)
            {
                return pressOff;
            }
        break;
        case 0b01:
            if (onBit)
            {
                return longPressOn;
            }
            if (offBit)
            {
                return longPressOff;
            }
        break;
        case 0b10:
            if (onBit)
            {
                return pressOnRepeat;
            }
            if (offBit)
            {
                return pressOffRepeat;
            }
        break;
        case 0b11:
            if (onBit)
            {
                return longPressOnRepeat;
            }
            if (offBit)
            {
                return longPressOffRepeat;
            }
            return longPressEnd;
        break;
        default:
            return noTask;
        break;
    }
    return noTask;
}

/// @brief Function to convert the raw bit on the bus to the 'telegram' structure
/// @param telegramRaw Pointer to the raw data
/// @param telegram Pointer to the (existing!) telegram structure.
void funkBusTelegramRawToTelegram(const funkBusTelegramRaw * telegramRaw, funkBusTelegram *telegram)
{
    telegram->button.serial = getSerialInfunkBustelegramRaw(telegramRaw);
    telegram->channel = getChannelTypeInFunkBustelegramRaw(telegramRaw);
    telegram->button.key = getKeyInFunkBustelegramRaw(telegramRaw);
    uint8_t sequence = getSequenceInFunkBustelegramRaw(telegramRaw);
    bool onBit = getOnBitInFunkBustelegramRaw(telegramRaw);
    bool offBit = getOffBitInFunkBustelegramRaw(telegramRaw);
    telegram->type = getTaskType(onBit, offBit, sequence);
}

/// @brief Queue all telegram bits to be send
/// @param telegramRaw Pointer to the raw-telegram representing the data to be send. 
/// @return True if all bits could be queued
bool funkBusQueuetelegramRaw(funkBusTelegramRaw * telegramRaw)
{
    xQueueReset(funkBusBitSendQueue);
    uint8_t byteCounter = 0;
    uint8_t mask = 0x80;
    bool b;
    while (byteCounter < sizeof(telegramRaw->raw))
    {
        uint8_t d = telegramRaw->raw[byteCounter];
        b = ((d & mask) > 0);
        if (xQueueSendToBack(funkBusBitSendQueue, &b, 0) != pdTRUE)
        {
            return false;
        }
        mask = mask / 2;
        if (mask == 0)
        {
            mask = 0x80;
            byteCounter++;
        }
    }
    b = true;
    if (xQueueSendToBack(funkBusBitSendQueue, &b, 0) != pdTRUE)
    {
        return false;
    }
    if (xQueueSendToBack(funkBusBitSendQueue, &b, 0) != pdTRUE)
    {
        return false;
    }
    return true;
}

// ---------------------------------------------------------------------------------------
// Call back function pointers
// ---------------------------------------------------------------------------------------

/// @brief callback function to process a valid received telegram.
funkBusReceiveCallBack cbFunkBusReceive = NULL;

// ---------------------------------------------------------------------------------------
// FunkBus handling
// ---------------------------------------------------------------------------------------

/// @brief Function to collect received bits from the bus connected to the receive-pin.
void funkBusReceiveTask()
{
    static funkBusTelegramRaw funkBusReceivetelegramRaw = {};
    static uint8_t bitMask = 0x80;
    static uint8_t byteCount = 0;
    static uint8_t shortBitCount = 0;
    static int64_t deltaTime;
    static bool bit1 = false;
    static bool bit0 = false;

    while (xQueueReceive(funkBusReceiveQueue, &deltaTime, 0) == pdTRUE)
    {
        bit0 = false;
        bit1 = false;
        bool restart = false;
        if ((deltaTime >= BIT1_MIN) && (deltaTime <= BIT1_MAX))
        {
            shortBitCount++;
            if (shortBitCount > 1)
            {
                shortBitCount = 0;
                bit1 = true;
            }
        }
        if ((deltaTime >= BIT0_MIN) && (deltaTime <= BIT0_MAX))
        {
            bit0 = true;
        }
        if (deltaTime > BIT0_MAX)
        {
            restart = true;
        }
        if (bit0 || bit1)
        {
            if (bit1)
            {
                if (byteCount < TELEGRAM_BYTECOUNT)
                {
                    funkBusReceivetelegramRaw.raw[byteCount] = funkBusReceivetelegramRaw.raw[byteCount] | bitMask;
                }
            }
            bitMask = bitMask >> 1;
            if (bitMask == 0)
            {
                bitMask = 0x80;
                byteCount++;
            }
            if (byteCount >= TELEGRAM_BYTECOUNT)
            {
                if (not funkBusChecktelegramRaw(&funkBusReceivetelegramRaw.raw[0]))
                {
                    ESP_LOGE(TAG, "Checksum error");    
                }
                else
                {
                    if (not funkBusCalcDataCheckBit(&funkBusReceivetelegramRaw))
                    {
                        ESP_LOGE(TAG, "Data parity error");    
                    }
                    else
                    {
                        if (cbFunkBusReceive != NULL)
                        {
                            funkBusTelegram telegram = {};
                            funkBusTelegramRawToTelegram(&funkBusReceivetelegramRaw, &telegram);
                            (*cbFunkBusReceive)(&telegram);
                        }
                    }
                }
                restart = true;
            }
        }
        if (restart)
        {
            funkBusCleartelegramRaw(&funkBusReceivetelegramRaw);
            bitMask = 0x80;
            byteCount = 0;
            shortBitCount = 0;
        }
    }
}

// Half bit time set to 512us (This is assumed by observing some Gira transmitters, uncertaincy is about +-15us)
#define FUNKBUS_TIMER_PERIOD    512 //us
// Arbitrary choosen period to repeatedly serve the task.
#define FUNKBUS_TASK_LOOP_PERIOD 10
// Next telegram time is 250ms.  (This is assumed by observing some Gira transmitters, uncertaincy is +-10ms)
#define FUNKBUS_SEND_DELAY (250 / FUNKBUS_TASK_LOOP_PERIOD)

/// @brief Send previously queued telegrams 
void funkBusSendTask()
{
    static uint8_t delaySendCounter = FUNKBUS_SEND_DELAY;
    if (delaySendCounter == 0)
    {
        if (uxQueueMessagesWaiting(funkBusBitSendQueue) == 0) // if previous transmission has finished
        {
            funkBusTelegramRaw funkBusSendtelegramRaw = {};
            if (xQueuePeek(funkBustelegramRawSendQueue, &funkBusSendtelegramRaw, 0) == pdTRUE)
            {
                if (not funkBusQueuetelegramRaw(&funkBusSendtelegramRaw))
                {
                    ESP_LOGE(TAG, "Could not queue telegramRaw");
                }
                else
                {
                    delaySendCounter = FUNKBUS_SEND_DELAY;                    
                    if (xQueueReceive(funkBustelegramRawSendQueue, &funkBusSendtelegramRaw, 0) != pdTRUE)
                    {
                        ESP_LOGE(TAG, "Unexpected empty queue");
                    }
                    else
                    {
                        if (esp_timer_start_periodic(funkBusBitSentTimer, FUNKBUS_TIMER_PERIOD) != ESP_OK)
                        {
                            ESP_LOGE(TAG, "Cannot start timer");
                        }
                    }
                }
            }
        }
    }
    if (delaySendCounter > 0)
    {
        delaySendCounter--;
    }
}

/// @brief The repeted called function by a free-rtos task to handle all funkbus functions
/// @param pvParameters 
void funkBusTask(void *pvParameters)
{
    while (true)
    {
        funkBusReceiveTask();
        funkBusSendTask();
        vTaskDelay(pdMS_TO_TICKS(FUNKBUS_TASK_LOOP_PERIOD));
    }
}

// task characteristics
// Due to the use of interrupt handling an queue-ing the priority can be lower.
// However, a delay of a telegram-length (about 50ms) is too long.
#define FUNKBUS_TASKS_STACKSIZE     4000
#define FUNKBUS_TASKS_TASK_PRIORITY 1

TaskHandle_t handleFunkBusTask = NULL;

/// @brief Create the nessecary task to reeive telegrams and send queued telegrams in a timely way.
void createFunkBusTask()
{
    if (xTaskCreate(funkBusTask, "FunkBusTask", FUNKBUS_TASKS_STACKSIZE, NULL, FUNKBUS_TASKS_TASK_PRIORITY, &handleFunkBusTask) != pdPASS)
    {
        ESP_LOGE(TAG, "Could not start funkbus task");
    }
}

/// @brief Kill the used free-rtos task if initialised
void killFunkBusTask()
{
    if (handleFunkBusTask != NULL)
    {
        vTaskDelete(handleFunkBusTask);
    }
}

// ---------------------------------------------------------------------------------------
// Public functions
// ---------------------------------------------------------------------------------------

/// @brief Call this function to intitialise the hardware en software
/// @param receivePin Pin number to fysical connect to the Gira-Receiver-Signal
/// @param sendPin Pin number to fysical connecton to the Gira actuators
/// @param cb Callback function is called on recepion of a valid telegram
void initFunkBus( int receivePin, int sendPin, funkBusReceiveCallBack cb)
{
    cbFunkBusReceive = cb;
    rxPin = (gpio_num_t)receivePin;
    txPin = (gpio_num_t)sendPin;
    setFunkBusSendHardware();
    setFunkBusReceiveHardware();
    createFunkBusTask();
}

/// @brief Call this function to stop the software and hardware.
void directStopFunkBus()
{
    ESP_LOGW(TAG, "Direct stop FunkBus");
    cbFunkBusReceive = NULL;
    gpio_isr_handler_remove(rxPin);
    esp_timer_stop(funkBusBitSentTimer);
    killFunkBusTask();
    gpio_set_level(txPin, SEND_PIN_RELEASE);
}

/// @brief Funtion to queue a telegram. Queued telegrams are timely send using the send-pin
/// @param telegram Initialised structure representing the data to be send.
/// @return True if telegram was succesfully queued.
bool sendFunkBusTelegram(const funkBusTelegram * telegram)
{
    funkBusTelegramRaw funkBusSendtelegramRaw = {}; 
    funkBusTelegramToTelegramRaw(telegram, &funkBusSendtelegramRaw);
    if (xQueueSendToBack(funkBustelegramRawSendQueue, &funkBusSendtelegramRaw, 0) != pdTRUE)
    {
        ESP_LOGE(TAG, "Could not queue telegram");
        return false;
    }
    return true;
}

/// @brief Function to get readable text representing a specific task
/// @param task Task to get a readable text
/// @return Pointer to a null-terminated string with the readable text. "" is returned if a task is not recognised.
char * funkBusTaskToChar(const telegramType task)
{
    switch (task)
    {
        case noTask:
            return (char *)"No task";
        break;
        case pressOn:
            return  (char *)"Press ON";
        break;
        case pressOff:
            return  (char *)"Press OFF";
        break;
        case longPressOn:
            return  (char *)"Long press ON";
        break;
        case longPressOff:
            return  (char *)"Long press OFF";
        break;
        case pressOnRepeat:
            return  (char *)"Press ON repeat";
        break;
        case pressOffRepeat:
            return (char *)"Press OFF repeat";
        break;
        case longPressOnRepeat:
            return  (char *)"Long press ON repeat";
        break;
        case longPressOffRepeat:
            return  (char *)"Long press OFF repeat";
        break;
        case longPressEnd:
            return  (char *)"Long press end";
        break;
    }
    return (char *)"";
}

/// @brief Determen a task is a non repeating task
/// @param task Task type to evaluate
/// @return true when a task is a repeat task
bool funkBusTaskIsRepeat(const telegramType task)
{
    switch (task)
    {
        case pressOnRepeat:
        case pressOffRepeat:
        case longPressOnRepeat:
        case longPressOffRepeat:
            return true;
        break;
        default:
            return  false;
        break;
    }
    return false;
}

/// @brief Function to detemen if a task is a button press. 
/// @param task Task type to evaluate
/// @return True if a task is a button press. False typically on automaticly added telegrams with repeat/end tasks to overcome lost messages in the rf-path.
bool funkBusTaskIsButtonPress(const telegramType task)
{
    switch (task)
    {
        case pressOn:
        case pressOff:
        case longPressOn:
        case longPressOff:
            return true;
        break;
        default:
            return  false;
        break;
    }
    return false;
}

/// @brief Function to display telegram content in a readable format.
/// @param title A title to add for convenience.
/// @param t Pointer to a telegram to be displayed
void printFunkBusTelegram(const char * tag, const char * title, const funkBusTelegram * t)
{
    funkBusTelegramRaw r = {}; 
    funkBusTelegramToTelegramRaw(t, &r);
    char * typeName = funkBusTaskToChar(t->type);
    ESP_LOGI(tag, "%s: Raw:%02X-%02X-%02X-%02X-%02X-%02X channel:%u serial:%u key:%u task:%s", title, r.raw[0], r.raw[1], r.raw[2], r.raw[3], r.raw[4],r.raw[5], t->channel, t->button.serial, t->button.key, typeName);
}

// ---------------------------------------------------------------------------------------
// end
// ---------------------------------------------------------------------------------------

