// ---------------------------------------------------------------------------------------

// Gira Funkbus Interface by EL
// Software to capture and send Funkbus messages
// see file : FunkBus protocol Berker.txt
// Hardware interface, see pictures of schematics and experimental hardware
// Version 01-05-2024

// ---------------------------------------------------------------------------------------

#pragma once

#ifndef funkBusTranseiver_H 
#define funkBusTranseiver_H

#ifdef __cplusplus
extern "C" {
#endif

#include <stdio.h>

// ---------------------------------------------------------------------------------------
// Examples to use this lib
// ---------------------------------------------------------------------------------------

/*
void funkBusReceiveCallBackHandler(funkBusTelegram * telegram)
{
    printFunkBusTelegram("Recv:", telegram);
}

#define FUNKBUS_SEND_PIN        ((gpio_num_t)GPIO_NUM_18)
#define FUNKBUS_RECEIVE_PIN     ((gpio_num_t)GPIO_NUM_14)

initFunkBus(FUNKBUS_RECEIVE_PIN, FUNKBUS_SEND_PIN, funkBusReceiveCallBackHandler);

funkBusTelegram telegram;
telegram.task = pressOn;
telegram.key = 2;
telegram.channel = channelA;
telegram.serial = 0x79FA;
printFunkBusTelegram("Send:", &telegram);
sendFunkBusTelegram(telegram);

See also handling in EL_FunkbusEvents
*/

// ---------------------------------------------------------------------------------------
// FunkBus specific structs
// ---------------------------------------------------------------------------------------

/// @brief Tasks are a combination of telegrams. Telegrams are of one of these types.
enum telegramType
{
    noTask = 0,
    pressOn,
    pressOff,
    longPressOn,
    longPressOff,
    pressOnRepeat,
    pressOffRepeat,
    longPressOnRepeat,
    longPressOffRepeat,
    longPressEnd
};

/// @brief Telegrams are about a task for a certain channel
///        Channel 'LS' is not tested.
enum funkBusChannel
{
    channelA = 0,
    channelB,
    channelC,
    channelLS,
};

/// @brief Telegrams in a task are about a certain button.
///        The button is identifyed by a serial nuber of the used transmitter and 
///        the button position of that number.
typedef struct
{
    uint16_t serial;
    uint8_t key;
} funkBusButton;


/// @brief Transmitted telegrams are always of this structure.
typedef struct
{
    funkBusChannel channel;
    funkBusButton button;
    telegramType type;
} funkBusTelegram;

// ---------------------------------------------------------------------------------------
// FunkBus handy defs
// ---------------------------------------------------------------------------------------

#define FUNKBUS_NO_KEY      0
#define FUNKBUS_NO_SERIAL   0

// ---------------------------------------------------------------------------------------
// FunkBus callback functiontypes
// ---------------------------------------------------------------------------------------

/// @brief Type of callback function to handle a received telegram.
///        Handling is not alowed to take more than 50ms.
typedef void (*funkBusReceiveCallBack)(funkBusTelegram * telegram);

// ---------------------------------------------------------------------------------------
// Public functions
// ---------------------------------------------------------------------------------------

// Lower level and library function control

void initFunkBus(int receivePin, int sendPin, funkBusReceiveCallBack cb);

bool sendFunkBusTelegram(const funkBusTelegram * telegram);

char * funkBusTaskToChar(const telegramType task);
bool funkBusTaskIsRepeat(const telegramType task);
bool funkBusTaskIsButtonPress(const telegramType task);

void printFunkBusTelegram(const char * tag, const char * title, const funkBusTelegram * telegram);

void directStopFunkBus();

// ---------------------------------------------------------------------------------------
// End
// ---------------------------------------------------------------------------------------

#ifdef __cplusplus
} /*extern "C"*/
#endif

#endif /*funkBusTranseiver_H*/


