// ---------------------------------------------------------------------------------------

// Gira Funkbus Events by EL
// Software to capture funcbus events
// Layer on receiving en sending telegrams using EL_FunkBus
// Version 01-05-2024

// ---------------------------------------------------------------------------------------

#include <stdio.h>
#include <cstdio>
#include <string.h>

#include "EL_FunkBusEvents.hpp"
#include "EL_FunkBus.hpp"

// ---------------------------------------------------------------------------------------
// Debug help
// ---------------------------------------------------------------------------------------

#include "esp_err.h"
#include "esp_log.h"
static const char *TAG = "FunkBusEvents";

// ---------------------------------------------------------------------------------------
// TelegramHistory table
// ---------------------------------------------------------------------------------------

 /*
 Funkbus is transporting telegrams
 Every action consists of a series of telegrams.
 The nature of RF-transmissions is tha a telegram can get lost.
 The series of telegrams alwos to lose some of them.
 Series of telegrams are send is a timely order.
 A series of telegram is called an action.
 An action is a set of telegrams. A 'telegram-set'.
 */

#define TELEGRAM_REPETION_TIME  250000   // 250ms expressed in micro seconds
#define TELEGRAM_SET_TTL        (4 * TELEGRAM_REPETION_TIME)

#define NO_TIMESTAMP            0;

enum funkBusState
{
    stateRest = 0,
    statePressOnReceived,
    statePressOffReceived,
    stateLongOnReceived,
    stateLongOffReceived,
    stateLongEndReceived,
};

typedef struct
{
    int64_t lastTelegramTimeStamp;
    telegramType task;
    funkBusButton button;
    uint8_t RepeatCount;
} typeTelegramHistory;

typeTelegramHistory telegramHistory;

// ---------------------------------------------------------------------------------------
// Memory for callback functions
// ---------------------------------------------------------------------------------------

funkBusEventCallBack eventCallBack = NULL;

// ---------------------------------------------------------------------------------------
// Internal functions
// ---------------------------------------------------------------------------------------

/// @brief Set the telegram history to a default, cleared state.
void clearTelegramHistory()
{
    telegramHistory.lastTelegramTimeStamp = NO_TIMESTAMP;
    telegramHistory.task = noTask;
    telegramHistory.button.key = FUNKBUS_NO_KEY;
    telegramHistory.button.serial = FUNKBUS_NO_SERIAL;
    telegramHistory.RepeatCount = 0;
}

// ---------------------------------------------------------------------------------------
// process received funkbus telegrams
// ---------------------------------------------------------------------------------------

#define DIM_MAX_REPEATE 8   // --> max 5x repeat, to prevent buffer overflows if users keep on transmitting repeats.

void doEventCallBack(funkBusEvent * event)
{
    if (eventCallBack != NULL)
    {
        (*eventCallBack)(event);
    }
}

/// @brief Handling received telegrams
/// @param telegram Pointer to the received telegram.
///                 After the call data the pointer points to gets lost.
void funkBusReceiveCallBackHandler(funkBusTelegram * telegram)
{
    // show in monitor
    printFunkBusTelegram(TAG, "Receive telegram", telegram);
    // set in any case the last timestamp
    int64_t timestamp = esp_timer_get_time();
    if (abs(timestamp - telegramHistory.lastTelegramTimeStamp) > TELEGRAM_SET_TTL)
    {
        ESP_LOGI(TAG, "Restart reception.");
        telegramHistory.task = noTask;
        telegramHistory.RepeatCount = 0;
    }
    // Prepeare event
    funkBusEvent event = {fbNoEvent, {FUNKBUS_NO_SERIAL, FUNKBUS_NO_KEY}, channelA};
    event.button = telegram->button;
    event.channel = telegram->channel;
    switch (telegram->type)
    {
        case pressOn:
        break;
        case pressOnRepeat:
            if ((telegramHistory.task == noTask) || (telegramHistory.task == pressOn))
            {
                event.event = fbSwitchOn;
                doEventCallBack(&event);
            }
        break;    
        case pressOff:
        break;
        case pressOffRepeat:
            if ((telegramHistory.task == noTask) || (telegramHistory.task == pressOff))
            {
                event.event = fbSwitchOff;
                doEventCallBack(&event);
            }
        break;    
        case longPressOn:
            if ((telegramHistory.task == noTask) || (telegramHistory.task == pressOn))
            {
                event.event = fbDimMore;
                doEventCallBack(&event);
            }
        break;
        case longPressOnRepeat:
            if ((telegramHistory.task == noTask) || (telegramHistory.task == longPressOn))
            {
                event.event = fbDimMoreRepeat;
                doEventCallBack(&event);
            }
            if (telegramHistory.task == longPressOnRepeat)
            {
                if (telegramHistory.RepeatCount < DIM_MAX_REPEATE)
                {
                    if ((telegramHistory.RepeatCount & 1) == 1)
                    {
                        event.event = fbDimMoreRepeat;
                        doEventCallBack(&event);
                    }
                    telegramHistory.RepeatCount++;
                }
            }
        break;    
        case longPressOff:
            if ((telegramHistory.task == noTask) || (telegramHistory.task == pressOff))
            {
                event.event = fbDimLess;
                doEventCallBack(&event);
            }
        break;
        case longPressOffRepeat:
            if ((telegramHistory.task == noTask) || (telegramHistory.task == longPressOff))
            {
                event.event = fbDimLessRepeat;
                doEventCallBack(&event);
            }
            if (telegramHistory.task == longPressOffRepeat)
            {
                if (telegramHistory.RepeatCount < DIM_MAX_REPEATE)
                {
                    if ((telegramHistory.RepeatCount & 1) == 1)
                    {
                        event.event = fbDimLessRepeat;
                        doEventCallBack(&event);
                    }
                    telegramHistory.RepeatCount++;
                }
            }
        break;
        case longPressEnd:
            if ((telegramHistory.task != longPressEnd))
            {
                event.event = fbDimEnd;
                doEventCallBack(&event);
            }
        break;    
        case noTask:
        break;
    }
    telegramHistory.task = telegram->type;
    telegramHistory.lastTelegramTimeStamp = timestamp;
}

// ---------------------------------------------------------------------------------------
// FunkBus send functions
// ---------------------------------------------------------------------------------------

/// @brief Function to do a wildcard switch (Not tested)
/// @param setOn 
/// @return True if the telegrams could be queued.
bool funkBusSwitchAll(const bool setOn)
{
    funkBusTelegram telegram;
    (setOn) ? telegram = {channelLS, {FUNKBUS_NO_SERIAL, FUNKBUS_NO_KEY}, longPressOn} : telegram = {channelLS, {FUNKBUS_NO_SERIAL, FUNKBUS_NO_KEY}, longPressOn};
    if (not sendFunkBusTelegram(&telegram)) {return false;}
    (setOn) ? telegram.type = longPressOnRepeat : telegram.type = longPressOffRepeat;
    if (not sendFunkBusTelegram(&telegram)) {return false;}
    if (not sendFunkBusTelegram(&telegram)) {return false;}
    if (not sendFunkBusTelegram(&telegram)) {return false;}
    return true;
}

/// @brief Function to send a telegrams repersenting switch-press.
/// @param on True emulates the on-switch, false emulates the off-switch
/// @param button Pointer to structure indicating the button
/// @param channel Pointer to a channel to use.
/// @return True if the telegram could be queued.
bool funkBusPress(const bool on, const funkBusButton * button, const funkBusChannel * channel)
{
    telegramType type;
    (on) ? type = pressOn : type = pressOff;
    funkBusTelegram telegram = {*channel, *button, type};
    if (not sendFunkBusTelegram(&telegram)) {return false;}
    return true;
}

/// @brief Function to send a set of telegrams to indicate longpress.
/// @param moreOn True causes to dim one step towards on, false couses to dim one step towards off. 
/// @param button Pointer to structure indicating the button
/// @param channel Pointer to a channel to use.
/// @return True if the telegrams could be queued.
bool funkBusLongPress(const bool on, const funkBusButton * button, const funkBusChannel * channel)
{
    telegramType type;
    (on) ? type = longPressOn : type = longPressEnd;
    funkBusTelegram telegram = {*channel, *button, type};
    if (not sendFunkBusTelegram(&telegram)) {return false;}
    if (not sendFunkBusTelegram(&telegram)) {return false;}
    return true;
}

/// @brief Function to send a set of telegrams to dim (lights) on a actuator using the send-pin.
/// @param moreOn True causes to dim one step towards on, false couses to dim one step towards off. 
/// @param button Pointer to structure indicating the button
/// @param channel Pointer to a channel to use.
/// @return True if the telegrams could be queued.
bool funkBusDim(const bool moreOn, const funkBusButton * button, const funkBusChannel * channel)
{
    if (not funkBusPress(moreOn, button, channel)) {return false;};
    funkBusTelegram telegram;
    (moreOn) ? telegram = {*channel, *button, longPressOn} : telegram = {*channel, *button, longPressOff};
    if (not sendFunkBusTelegram(&telegram)) {return false;}
    if (not sendFunkBusTelegram(&telegram)) {return false;}
    return true;
}

/// @brief Function to send a set of telegrams to repeat dim (lights) on a actuator using the send-pin.
/// @param moreOn True causes to dim one step towards on, false couses to dim one step towards off. 
/// @param button Pointer to structure indicating the button
/// @param channel Pointer to a channel to use.
/// @return True if the telegrams could be queued.
bool funkBusDimRepeat(const bool moreOn, const funkBusButton * button, const funkBusChannel * channel)
{
    funkBusTelegram telegram;
    (moreOn) ? telegram = {*channel, *button, longPressOnRepeat} : telegram = {*channel, *button, longPressOffRepeat};
    if (not sendFunkBusTelegram(&telegram)) {return false;}
    if (not sendFunkBusTelegram(&telegram)) {return false;}
    return true;
}

/// @brief Function to send a set of telegrams to repeat dim (lights) on a actuator using the send-pin.
////       The first LongPressEnd types telegram should be send after 0.13s. This requirement is not met
///        as all telegrams are send every 0.25seconds.
/// @param moreOn True causes to dim one step towards on, false couses to dim one step towards off. 
/// @param button Pointer to structure indicating the button
/// @param channel Pointer to a channel to use.
/// @return True if the telegrams could be queued.
bool funkBusDimEnd(const funkBusButton * button, const funkBusChannel * channel)
{
    funkBusTelegram telegram = {*channel, *button, longPressEnd};
    if (not sendFunkBusTelegram(&telegram)) {return false;}
    if (not sendFunkBusTelegram(&telegram)) {return false;}
    return true;
}

// ---------------------------------------------------------------------------------------
// Public functions
// ---------------------------------------------------------------------------------------

/// @brief Initialise layer FunkBusEvents and layer FunkBus
/// @param receivePin pin to fysical receiving funkbus low-level protocol
/// @param sendPin pin to fysical receiving funkbus low-level protocol
/// @param eventCb Optional function to set to be called on receiving an event.
void initFunkBusEvents(const int receivePin, const int sendPin, const funkBusEventCallBack eventCb)
{
    clearTelegramHistory();
    eventCallBack = eventCb;
    initFunkBus(receivePin, sendPin, funkBusReceiveCallBackHandler);
}

/// @brief Convert the event to telegrams and send them on the FunkBus.
///        Using this function on received events make the module transparant.
/// @param event Event to be send
/// @return true if event could be comletely queued on the send buffer
bool sendFunkBusEvent(const funkBusEvent * event)
{
    switch (event->event)
    {
        case fbNoEvent:
            return false;
        break;
        case fbSwitchOn:
            return funkBusSwitch(true, &event->button, &event->channel);
        break;
        case fbSwitchOff:
            return funkBusSwitch(false, &event->button, &event->channel);
        break;
        case fbDimMore:
            return funkBusDim(true, &event->button, &event->channel);
        break;
        case fbDimMoreRepeat:
            return funkBusDimRepeat(true, &event->button, &event->channel);
        break;
        case fbDimLess:
            return funkBusDim(false, &event->button, &event->channel);
        break;
        case fbDimLessRepeat:
            return funkBusDimRepeat(false, &event->button, &event->channel);
        break;
        case fbDimEnd:
            return funkBusDimEnd(&event->button, &event->channel);
        break;
    }
    return false;
}

/// @brief Function to perform on the funkbas a full set of telegrams to dim on step
/// @param more True causes to dim one step towards on, false couses to dim one step towards off.
/// @param button Pointer to structure indicating the button
/// @param channel Pointer to a channel to use.
/// @return True if the telegrams could be queued.
bool funkBusDimStep(const bool more, const funkBusButton * button, const funkBusChannel * channel)
{
    if (not funkBusPress(more, button, channel)) {return false;}
    if (not funkBusDim(more, button, channel)) {return false;}
    if (not funkBusDim(more, button, channel)) {return false;}
    if (not funkBusDimEnd(button, channel)) {return false;}
    return true;
}

/// @brief Function to send a full set of telegrams to switch (lights) on a actuator using the send-pin.
/// @param setOn True causes to switch on
/// @param button Pointer to structure indicating the button
/// @param channel Pointer to a channel to use.
/// @return True if the telegrams could be queued.
bool funkBusSwitch(const bool on, const funkBusButton * button, const funkBusChannel *channel)
{
    if (not funkBusPress(on, button, channel)) {return false;}
    funkBusTelegram telegram;
    (on) ? telegram = {*channel, *button, pressOnRepeat} : telegram = {*channel, *button, pressOffRepeat};
    if (not sendFunkBusTelegram(&telegram)) {return false;}
    if (not sendFunkBusTelegram(&telegram)) {return false;}
    if (not sendFunkBusTelegram(&telegram)) {return false;}
    return true;
}

/// @brief Function to express an event-type in readable text.
///        It's c++ so all options are coded.
/// @param eventType The event to express
/// @return pointer to the hardcoded null-terminated string.
char * funkBusEventTypeToStr(const funkBusEventType eventType)
{
    switch (eventType)
    {
        case fbNoEvent:
            return (char *)"No event";
        break;
        case fbSwitchOn:
            return (char *)"Switch on";
        break;
        case fbSwitchOff:
            return (char *)"Switch off";
        break;
        case fbDimMore:
            return (char *)"Dim more";
        break;
        case fbDimLess:
            return (char *)"Dim less";
        break;
        case fbDimMoreRepeat:
            return (char *)"Dim more repeat";
        break;
        case fbDimLessRepeat:
            return (char *)"Dim less repeat";
        break;
        case fbDimEnd:
             return (char *)"Dim end";
        break;
    }
    return (char *)"";
}

/// @brief Log a readable event.
/// @param tag Use tag to make the function apear it is genereted fron a certain source area.
/// @param title Use title to inform about the source location the log is done
/// @param e Pointer to the event to be displayed.
void printFunkBusEvent(const char * tag, const char * title, const funkBusEvent * e)
{
    char * eventType = funkBusEventTypeToStr(e->event);
    ESP_LOGI(tag, "%s: channel:%u serial:%u key:%u event:%s", title, e->channel, e->button.serial, e->button.key, eventType);
}


// ---------------------------------------------------------------------------------------
// end
// ---------------------------------------------------------------------------------------

